/* KALACAKRA CALENDAR CALCULATOR, K1.C */

/*********************************************************************************
License for Kcal - Kalacakra calendar software

Copyright (c) 2011 Edward Henning

Permission is hereby granted, free of charge, to any person  obtaining a copy of
this software and associated documentation files (the "Software"), to deal in the
Software without restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
Software, and to permit persons to whom the Software is furnished to do so, subject
to the following conditions:

The above copyright notice and this permission notice shall be included in all copies
or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
************************************************************************************/

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <string.h>
#include <dos.h>
#include <math.h>
#include "jd.h"

int chk_eclipse ( double, int );
double jul_dat ( int, int, int );
void sun_mon ( double );
double modfeh ( double, double );
void chg_tim ( int, int, int, int, int );
void prn_cal ( void );
int chk_spec ( long int, long int);
void chk_sadag ( long int, long int );
void get_new_moon ( double *, double *, double, double );
double get_tithi ( double, double );
int get_NFMoon ( double, int );
int chk_sign ( double, double );
double get_deltaT ( double );
double getmeanSun ( double );
int getphrod ( int, int, double );
void conv2hms ( int*, int*, int*, int, int, int );

char filnam[14];
char ch_num[8];

char eclps_str[80];

int inc_chin = 1; // Flag to include Chinese information.
int d_bug = 0;
char cal_site[50];
double site_long;
double jul_factor;
double jul_SC, jul_NFM; // Values for time of sign change and new/full Moon.

int  d_d, m_m, y_y; // Globals, because needed by get_deltaT ()
int  target_year;
int  first_Cmnth_flag, first_Kmnth_flag, last_mnth_flag;
long int juldat;
int yan_kwong_flg = 0;
int zin_phung_flg = 0;

struct new_moon_details
  {
    int dd;
    int mm;
    int yy;
    int Zsignstart;    // Range: 0-11
    int Zsignend;      // Range: 0-11
    int mnthnumK;      // Range: 1-12 - Kalacakra month number: 1 = nag zla
    int intercal;
  } newmon[20];   // Maximum index reached in testing = 17

extern struct eclipse_details
  {
    int partial;
    int annular;
    int total;
    float mag;
  } eclps;

extern double mol, sol;

double  juldatf; // Julian day
long int   dayofweek;   // Day of the week

char *dayoweek[7] = { "Sat", "Sun", "Mon", "Tue", "Wed", "Thu", "Fri" };
char *wdayelem[7] = { "Earth", "Fire", "Water", "Fire", "Water", "Wind", "Earth" };

char *lunmanP[27];
char *lunmanT[27] = { "tha skar", "bra nye", "smin drug", "snar ma", "mgo",
                    "lag", "nabs so", "rgyal", "skag", "mchu", "gre", "dbo",
                    "me bzhi", "nag pa", "sa ri", "sa ga", "lha mtshams",
                    "snron", "snrubs", "chu stod", "chu smad", "gro zhin",
                    "mon gre", "mon gru", "khrums stod", "khrums smad",
                    "nam gru" };

char *lmanelem[27] = { "Wind", "Fire", "Fire", "Earth", "Wind", "Water",
            "Wind", "Fire", "Water", "Fire", "Fire", "Wind", "Wind", "Wind",
            "Wind", "Wind", "Earth", "Earth", "Water", "Water", "Earth",
            "Earth", "Water", "Earth", "Fire", "Water", "Water" };

char *lunmanS[27] = { "Ashvini", "Bharani", "Krittika", "Rohini",
                      "Mrigashiras", "Ardra", "Punarvasu", "Pushya",
                      "Ashlesha", "Magha", "Purvaphalguni", "Uttaraphalguni",
                      "Hasta", "Citra", "Svati", "Vishakha", "Anuradha",
                      "Jyeshtha", "Mula", "Purvashadha", "Uttarashadha",
                      "Shravana", "Dhanishta", "Shatabhishaj",
                      "Purvabhadrapada", "Uttarabhadrapada", "Revati" };

char *yogaP[27];
char *yogaT[27] = { "rnam sel", "mdza' bo", "tshe dang ldan pa", "skal bzang",
                  "dge byed", "shin tu 'grams", "las bzang", "'dzin byed",
                  "zug rngu", "'grams", "'phel", "brtan pa", "yongs bsnun",
                  "dga' ba", "rdo rje", "dngos grub", "phan tshun",
                  "mchog can", "yongs 'joms", "zhi ba", "grub pa",
                  "bsgrub bya", "dge ba", "dkar po", "tshangs pa",
                  "dbang po", "'khon 'dzin" };

char *yogaS[27] = { "Vishkambha", "Priti", "Ayushmat", "Saubhagya",
                    "Shobhana", "Atiganda", "Sukarma", "Dhriti", "Shula",
                    "Ganda", "Vriddhi", "Dhruva", "Vyaghata", "Harshana",
                    "Vajra", "Siddhi", "Vyatipata", "Variyas", "Parigha",
                    "Shiva", "Siddha", "Sadhya", "Shubha", "Shukla",
                    "Brahman", "Indra", "Vaidhriti" };

char *byedP[11];
char *byedT[11] = { "gdab pa", "byis pa", "rigs can", "til rdung",
                  "khyim skyes", "tshong ba", "vii", "mi sdug pa",
                  "bkra shis", "rkang bzhi", "klu" };

char *byedS[11] = { "Vava", "Valava", "Kaulava", "Taitila", "Garaja", "Vanija",
                    "Vishti", "Kintughna", "Shakuni", "Catushpada", "Naga" };

char *cycanim[12] = { "Rabbit", "Dragon", "Snake", "Horse", "Sheep", "Monkey",
                      "Bird", "Dog", "Pig", "Mouse", "Ox", "Tiger" };

char *cycelem[10] = { "Wood", "Fire", "Fire", "Earth", "Earth", "Iron",
                      "Iron", "Water", "Water", "Wood" };

char *cycelem1[5] = { "Fire", "Earth", "Iron", "Water", "Wood" };

char *cycgendE[2] = { "male", "female" };

char *cycpar[8] = { "Zon", "Li", "Khon", "Dwa", "Khen", "Kham", "Gin", "Zin"
                                                                         };
char *cycparT[8] = { "Zon", "Li", "Khon", "Dwa", "Khen", "Kham", "Gin", "Zin"
                                                                         };
char *cycparE[8] = { "Gentle", "Clinging", "Receptive", "Joyous", "Creative",
                     "Abysmal", "Keeping Still", "Arousing" };

char *wmonths[12] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug",
                      "Sep", "Oct", "Nov", "Dec" };

char *Zsigns[12] = { "Aries", "Taurus", "Gemini", "Cancer", "Leo", "Virgo",
                     "Libra", "Scorpio", "Sagittarius", "Capricorn",
                     "Aquarius", "Pisces" };

char *MnthS[12] = { "Caitra", "Vaishakha", "Jyeshta", "Ashadha",
                    "Shravana", "Bhadrapada", "Ashvina", "Kartikka",
                    "Margashirsha", "Pausha", "Magha", "Phalguna" };

char *M_season[12] = { "Mid-Spring", "Late-Spring", "Early-Summer",
                       "Mid-Summer", "Late-Summer", "Early-Autumn",
                       "Mid-Autumn", "Late-Autumn", "Early-Winter",
                       "Mid-Winter", "Late-Winter", "Early-Spring" };

char *Clunman[28] = { "Jiao", "Kang", "Di", "Fang", "Xin", "Wei", "Ji", "Dou",
                    "Niu", "Nu", "Xu", "Wei", "Shi", "Bi", "Kui", "Lou",
                    "Wei", "Mao", "Bi", "Zui", "Can", "Jing", "Gui", "Liu",
                    "Xing", "Zhang", "Yi", "Zhen" };

char *Prabhava[60] = { "Prabhava", "Vibhava", "Shukla", "Pramoda", "Prajapti",
                       "Anggiras", "Shrimukha", "Bhava", "Yuvika", "Dhritu",
                       "Ishvara", "Vahudhvanya", "Pramadi", "Vikrama",
                       "Brishabha", "Citra", "Bhanu", "Bhanutara", "Virthapa",
                       "Akshaya", "Sarvajit", "Sarvadhari", "Virodhi",
                       "Vikrita", "Khara", "Nanda", "Vijaya", "Jaya", "Mada",
                       "Durmukha", "Hemalambha", "Vilambhi", "Vikari",
                       "Sarvavati", "Slava", "Shubhakrita", "Shobhana",
                       "Krodhi", "Vishvabandhu", "Parabhava", "Pravamga",
                       "Kilaka", "Saumya", "Sadharana", "Virobhakrita",
                       "Paradhari", "Pramadi", "Ananda", "Rakshasa", "Anala",
                       "Vinggala", "Kaladuti", "Siddhartha", "Rudra",
                       "Durmati", "Dundubhi", "Rudhirura", "Raktakshi",
                       "Krodhana", "Kshayaka" };

char *phrodE[28] = { "joy", "time-club", "control", "humanity", "youth",
               "raven", "victory-banner", "eternal-knot", "vajra",
               "attainment", "parasol", "friend", "intellect", "desire",
               "firebrand", "emaciation", "death-lord", "arrow", "success",
               "lance", "nectar", "pestle", "elephant", "tigress",
               "exhaustion", "movement", "stability", "increase" };

// Chinese data:
long int c_genday;
int sd_clunmanx;
int ld_parx, sd_parx;
int ld_smex, sd_smex;
int ld_planx, ld_lunmx;
int ldo_animx, ldo_elemx, ldo_parx, ldo_smex;  // for omitted days.
int yr_gender;    // 1 = female, odd; 0 = male, even
int yr_animx, mt_animx, ld_animx, sd_animx;
int yr_elemx, mt_elemx, ld_elemx, sd_elemx;
int yr_smex;

FILE *fptgt;
FILE *fptgt_d; // For debug information
char outbuf[120];
char outbufSC[60];
char outbufNFM[50];

extern char  *e_str[];

void main (argc, argv)
    int argc;
    char *argv[];
  {
    int    finish, i, hit;
    double moon_elong, last_moon_elong;
    double NM_juldat, NM_sol, sol_hold;
    double NM_meansol; // Mean Sun at new Moon.
    int    mnth12_found, just_done_newmoon;
    int    first_new_year_found, second_new_year_found;
    int    intercal_found; // For testing only.

    if ( argc > 1 )
      {
        for ( i = 1; i < argc; ++i )
          {
            if ( i == 1 )  // file name
              strcpy ( filnam, argv[i] );
            else if ( i == 2 )  // site name
              strcpy ( cal_site, argv[i] );
            else if ( i == 3 )  // Year
              {
                strcpy ( ch_num, argv[i] );
                y_y = atoi ( ch_num );
                printf ( "Kalacakra calendar software. Version 2.1\n") ;
                printf ( "Target year = %d\n", y_y );
              }
            else if ( i == 4 )  // Longitude
              {
                strcpy ( ch_num, argv[i] );
                site_long = atof ( ch_num );
                printf ( "Longitude = %f\n", site_long );
              }
          } // END - for ( i = 1; i < argc; ++i )
        if ( cal_site[0] )
          {
            i = 0;
            do
              {
                if ( cal_site[i] == '_' )
                  cal_site[i] = 32;
                ++i;
              } while ( cal_site[i] > '\0' );
          }
      } // END - if ( argc > 1 )
    else
      {
        site_long = 0.0; // For Greenwich as default
        printf ("\n\nEnter the year required: ");
        scanf ("%d", &y_y );
        printf ("\n");
      }
    target_year = y_y;
    mnth12_found = 0;
    first_new_year_found = 0;
    second_new_year_found = 0;
    intercal_found = 0;  // TESTING ONLY

// Kalacakra New Year cannot start before about 22nd Feb, Chinese before 21
// Jan. Start searching from 1st Jan:

    d_d = 1;
    m_m = 1;
    last_moon_elong = 0.0;
    hit = 0;
    finish = 0;
    just_done_newmoon = 0;
    jul_factor = 5.0/24.0 - site_long / 360.0;  // for 5am LMST at site
    printf ( "YEAR = %d\n", y_y );

    if ( inc_chin ) // Calculate Chinese year data.
      {
        if ( y_y < 7 )
          {
            yr_animx = ( y_y + 1200 - 7 ) % 12;
            yr_elemx = ( ( y_y + 1200 ) / 2 - 3 ) % 5;
            yr_gender = ( y_y + 1200 ) % 2; // 1 = female, odd; 0
          }
        else
          {
            yr_animx = ( y_y - 7 ) % 12;
            yr_elemx = ( y_y / 2 - 3 ) % 5;
            yr_gender = y_y % 2; // 1 = female, odd; 0 = male
          }
        yr_smex = 9 - ( ( y_y + 7 ) % 9 );

        sprintf ( outbuf, "\nChinese New Year: %d, %s-%s-%s", y_y,
           cycelem1[ yr_elemx ], cycgendE[ yr_gender ],
           cycanim[ yr_animx ] );
        printf ( "%s\n", outbuf );
        printf ( "sme ba = %d\n", yr_smex );
      }

// First, go through year finding new moons and setting month definitions.
    while ( finish == 0 )
      {
        juldatf = jul_dat ( d_d, m_m, y_y ); // returns the x.5 value.
                                             // ie. beginning of UT day.
        juldatf = juldatf + jul_factor; // Add time zone factor.
// juldatf is now equivalent to 5am LMST on our target meridian
        sun_mon ( juldatf ); // This adds Delta T
        moon_elong = mol - sol;
        sol_hold = sol;  // Solar longitude at 5am LMST.
        if ( moon_elong < 0.0 )
          moon_elong = moon_elong + 360.0;
        if ( moon_elong < last_moon_elong ) // We have gone past a New Moon
          {
            newmon[hit].dd = d_d;  // First day after new Moon.
            newmon[hit].mm = m_m;
            newmon[hit].yy = y_y;

// We need to find exact time of New Moon:
            get_new_moon ( &NM_juldat, &NM_sol, juldatf, moon_elong );
            just_done_newmoon = 1;

// We store an index number for the solar zodiac sign at new Moon.

// Use this line if month definition is true Sun at true new Moon:
//          newmon[hit].Zsignstart = (int) floor ( NM_sol / 30.0 );

// Use these two lines if month definition is mean Sun at true new Moon:
// Should we calculate exact time of mean new moon - perhaps useful??
            NM_meansol = getmeanSun ( NM_juldat ); // This adds Delta T
            newmon[hit].Zsignstart = (int) floor ( NM_meansol / 30.0 );

// Show the user we are actually doing something:
            printf ( "Jul: %f, Mean Sun = %f, True Sun = %f\n",
                      NM_juldat, NM_meansol, NM_sol );

// Use this line if month definition is daybreak, true Sun:
//          newmon[hit].Zsignstart = (int) floor ( sol_hold / 30.0 );

            if ( hit > 0 ) // Not our first new Moon
              {
                newmon[hit-1].Zsignend = newmon[hit].Zsignstart;
// Do we have a month in which the Sun does not change sign?
                if ( newmon[hit-1].Zsignstart == newmon[hit-1].Zsignend )
                  {
//                  printf ( "Hit = %d\n", hit ); // Removed debug code
                    newmon[hit-1].mnthnumK = newmon[hit-1].Zsignend + 2;
                    if ( newmon[hit-1].mnthnumK > 12 )
                      newmon[hit-1].mnthnumK -= 12;
                    newmon[hit-1].intercal = 1;
                    printf ( "Intercalary month: %d\n", newmon[hit-1].mnthnumK );
                    if ( hit < 3 && newmon[hit-1].mnthnumK == 12 ) // FOR
                                                               // TESTING
                      mnth12_found = 1;
                    if ( intercal_found )  // TESTING ONLY
                      printf ( "TWO INTERCALARY MONTHS!!!\n" );
                    intercal_found = 1;
//                  getch ();
                  }
                else  // Normal month
                  {
                    newmon[hit-1].mnthnumK = newmon[hit-1].Zsignend + 1;
                    if ( newmon[hit-1].mnthnumK > 12 )
                      newmon[hit-1].mnthnumK -= 12;
                    newmon[hit-1].intercal = 0;

// Following two "if"s are for checking for possible problems:
                    if ( hit < 3 && newmon[hit-1].mnthnumK == 12 )
                      mnth12_found = 1;
                    if ( newmon[hit-1].Zsignend - newmon[hit-1].Zsignstart == 2
                                                                           )
                      {
                        printf ( "DOUBLE SIGN ENTRY: %d\n",
                                 newmon[hit-1].Zsignstart + 1 );
                        getch ();
                      }
                  }
              }
            if ( first_new_year_found == 0 )
              {
                if ( newmon[hit-1].mnthnumK == 1 )
                  {
                    first_Kmnth_flag = hit - 1;
                    first_new_year_found = 1;
                  }
// This to fix the missing of intercalary first months: 11/3/2008
// Point is: if we have already found the first month, do not mark a
// delayed month 1 as the first. Is there a similar problem with line above?
                else if ( newmon[hit-1].mnthnumK == 12 && hit > 1
                          && newmon[hit-2].mnthnumK != 12 )
                  {
                    first_Cmnth_flag = hit - 1;
//                  first_new_year_found = 1;
                  }
              }
            if ( second_new_year_found == 0 && y_y > target_year &&
                 newmon[hit-1].mnthnumK == 1 )
              {
                last_mnth_flag = hit - 2;
                second_new_year_found = 1;
                finish = 1;
              }
            ++hit;
          }
// Do another day:
        last_moon_elong = moon_elong;

        if ( just_done_newmoon == 1 )
          {
            chg_tim ( 0, 0, 26, 0, 0 );
            just_done_newmoon = 0;
          }
        else
          chg_tim ( 0, 0, 1, 0, 0 );
      } // End of loop searching for new Moons

    if ( !mnth12_found )  // Keep this as an error check for main program.
      {
        printf ( "Twelfth month not found\n" );
        getch ();
      }

    finish = 0;
    prn_cal ();
  } // END - main

// Check for special festival days, etc.
int chk_spec ( long int m, long int t )
  {
// First, convert to Tibetan month numbering.
// There are strong arguments for shifting along a month, as these results
// give, for example the Enlightenment of the Buddha in Vaishakha, but in this
// calendar this is the 3nd month, not the 4th in Tibetan counting.
    m = m + 2;
    if ( m > 12 )
      m = m - 12;

    switch ( m )
      {
        case 1:
          if ( t == 1L )
            {
              sprintf ( outbuf, "%s",
                "From 1st to 15th, Demonstration of Miracles." );
              return (1);
            }
        break;
        case 2:
          ;
        break;
        case 3:
          if ( t == 15L )
            {
              sprintf ( outbuf, "%s", "Revelation of the Kalacakra Tantra." );
              return (1);
            }
        break;
        case 4:
          if ( t == 8L )
            {
              sprintf ( outbuf, "%s", "Birth of the Buddha." );
              return (1);
            }
          else if ( t == 15L )
            {
              sprintf ( outbuf, "%s",
                "Enlightenment and Parinirvana of the Buddha." );
              return (1);
            }
        break;
        case 5:
          ;
        break;
        case 6:
          if ( t == 4L )
            {
              sprintf ( outbuf, "%s", "Turning of the Wheel of the Dharma." );
              return (1);
            }
          else if ( t == 15L )
            {
              sprintf ( outbuf, "%s",
                        "The Buddha's entry into the womb of his mother." );
              return (1);
            }
        break;
        case 8:
          ;
        break;
        case 9:
          if ( t == 22L )
            {
              sprintf ( outbuf, "%s",
                "Descent of the Buddha from the realm of the gods." );
              return (1);
            }
        break;
        case 10:
          ;
        break;
        case 11:
          ;
          break;
        case 12:
          ;
        break;
        default:
          ;
        break;
      }
    return (0);
  } // END - chk_spec ()

void chk_sadag ( long int m, long int t )
// Check for main Earth-lords, "sa bdag"
  {
// First, convert to Tibetan month numbering:  - again!
    m = m + 2;
    if ( m > 12 )
      m = m - 12;

// First, "yan kwong":

    yan_kwong_flg = 0;

    if ( m == 1L && t == 13L ) yan_kwong_flg = 1;
    else if ( m == 2L && t == 11L ) yan_kwong_flg = 1;
    else if ( m == 3L && t == 9L ) yan_kwong_flg = 1;
    else if ( m == 4L && t == 7L ) yan_kwong_flg = 1;
    else if ( m == 5L && t == 5L ) yan_kwong_flg = 1;
    else if ( m == 6L && t == 3L ) yan_kwong_flg = 1;
    else if ( m == 7L && t == 1L ) yan_kwong_flg = 1;
    else if ( m == 7L && t == 29L ) yan_kwong_flg = 1;
    else if ( m == 8L && t == 27L ) yan_kwong_flg = 1;
    else if ( m == 9L && t == 25L ) yan_kwong_flg = 1;
    else if ( m == 10L && t == 23L ) yan_kwong_flg = 1;
    else if ( m == 11L && t == 21L ) yan_kwong_flg = 1;
    else if ( m == 12L && t == 19L ) yan_kwong_flg = 1;

// Now, "zin phung":

    zin_phung_flg = 0;

    if ( m == 1L || m == 2L || m == 3L )
      {
        if ( t == 1 || t == 7 || t == 13 || t == 19 || t == 25 )
          zin_phung_flg = 1;
      }
    else if ( m == 4L || m == 5L || m == 6L )
      {
        if ( t == 6 || t == 12 || t == 18 || t == 24 || t == 30 )
          zin_phung_flg = 1;
      }
    else if ( m == 7L || m == 8L || m == 9L )
      {
        if ( t == 3 || t == 9 || t == 15 || t == 21 || t == 27 )
          zin_phung_flg = 1;
      }
    else if ( m == 10L || m == 11L || m == 12L )
      {
        if ( t == 4 || t == 10 || t == 16 || t == 22 || t == 28 )
          zin_phung_flg = 1;
      }

// dpyid gsum - stag spre - moves from south to north
// dbyar gsum - glang lug - moves from west to east
// ston gsum - khyi 'brug - moves from north to south
// dgun gsum - phag sbrul - moves from east to west
  }

void get_new_moon ( double * NM_jdat, double * NM_sol, double juldat,
                    double melong )
  {
    int done = 0;
    do
      {
        juldat = juldat - ( melong / 13.2 ); // Moon moves 13.2 degrees a day
        sun_mon ( juldat ); // This adds Delta T
        melong = mol - sol;
        if ( melong > -0.00001 && melong < 0.00001 )
          done = 1;
        if ( melong < -330.0 )
          melong = melong + 360.0;
        else if ( melong > 330.0 )
          melong = melong - 360.0;
        if ( melong < -15.0 || melong > 15.0 )
          {
            printf ( "ERROR - Sun: %f, Moon: %f\n", sol, mol );
            getch ();
          }
      } while ( !done );
    *NM_jdat = juldat;
    *NM_sol = sol;
  } // END - get_new_moon

double get_tithi ( double tgtelong, double juldat )
// Get time of lunar day change
  {
    int done = 0;
    double el_diff, jul_nxt, melong;
    jul_nxt = juldat;
    do
      {
        sun_mon ( jul_nxt ); // This adds Delta T
        melong = mol - sol;
        if ( melong < 0.0 )
          melong = melong + 360.0;
        el_diff = tgtelong - melong;
        if ( el_diff > 180.0 )
          el_diff = 360.0 - el_diff;
        else if ( el_diff < -180.0 )
          el_diff = 360.0 + el_diff;
        if ( el_diff > -0.00001 && el_diff < 0.00001 )
          done = 1;
        else
          jul_nxt = jul_nxt + ( el_diff / 13.2 ); // Moon moves 13.2 degrees a day
        if ( el_diff > 15.0 || el_diff < -15.0 )
          {
            printf ( "PROBLEM\n" );
            getch ();
          }
      } while ( !done );
    return ( jul_nxt );
  } // END - get_tithi

int chk_sign ( double juldat, double sl )
// juldat is set to juldatf = time at daybreak at target location.
// Checking for change of Sun sign.
  {
    double jd, tgt, dif;
    int s1, s2, n, p, b;
    int done = 0;

    s1 = (int) floor ( sl / 30.0 );
    jd = juldat + 1.0;
    sun_mon ( jd ); // This adds Delta T
    s2 = ( int ) floor ( sol / 30.0 );
    if ( s1 != s2 )
      {
        jd = juldat;
        tgt = (double) s2 * 30.0;
        if ( s2 == 0 )
          tgt = 360.0;
        dif = tgt - sl;
        do
          {
            jd = jd + dif; // Sun moves just less than 1 degree a day
            sun_mon ( jd );
            dif = tgt - sol;

            if ( dif > -0.00001 && dif < 0.00001 ) // Much less than a second.
              done = 1;
          } while ( !done );

        jul_SC = jd;
        dif = 60.0 * ( jd - juldat );  // Seems right.
        n = (int) floor ( dif );
        dif = 60.0 * ( dif - (double) n );
        p = (int) floor ( dif );
        dif = 6.0 * ( dif - (double) p );
        b = (int) floor ( dif );
        sprintf ( outbufSC, "\x20\x20Sun enters %s at %d:%d:%d.",
                          Zsigns[s2], n, p, b );
        return (1);
      }
    return (0);
  } // END - chk_sign ()

void prn_cal ( void )
  {
    int newyearintcal = 0;
    int i, more, start_month, Cmonth;
    int mth_ind;
    int tt, tt2, last_tt, tt_lhag, tt_chad;
    int eclipse_flg;
    int  lunmanx, lunmannext, sbyorba, byedpa, byedpax;
    int NFM_flag, SC_flag, lmchg;
    double x;
    double moon_elong, sol1, mol1, moon_elong1;
    double sol2, mol2, moon_elong2;
    double tt1chg, ott1chg, tt2chg, tt1tgt, tt2tgt;
    double lmchgjdatf, tgtmol, dif;
    int lmcn, lmcp, lmcb; // Lunar mansion change time details.
    int phrod1, phrod2;

    last_tt = 30;   // As if we had just done a month, for beginning of year.
    tt_lhag = 0;

// Comment out as required:

/****** For Sankrit names ********/
      for ( i = 0; i < 27; ++i )
        lunmanP[i] = lunmanS[i];
      for ( i = 0; i < 27; ++i )
        yogaP[i] = yogaS[i];
      for ( i = 0; i < 11; ++i )
        byedP[i] = byedS[i];
/********************/

/******* For Tibetan names ********
      for ( i = 0; i < 27; ++i )
        lunmanP[i] = lunmanT[i];
      for ( i = 0; i < 27; ++i )
        yogaP[i] = yogaT[i];
      for ( i = 0; i < 11; ++i )
        byedP[i] = byedT[i];
**************/

    if ( filnam[0] == '\0' )
      sprintf ( filnam, "%d.dat", target_year );
    fptgt = fopen ( filnam, "w" );
    if ( fptgt == NULL )
      {
        printf ( "Error opening target file!\n" );
        getch ();
        return;
      }
     i = target_year - 127; // Global - set before routine is called.
     i = i % 60;

     if ( cal_site[0] > 0 )
       {
         sprintf ( outbuf,
               "Calendar for %s, and other places near longitude %.2f",
                cal_site, site_long );
         fprintf ( fptgt, "%s\n\n", outbuf );
       }

// New year information
     if ( inc_chin )
       {
         sprintf ( outbuf, "Chinese New Year: %s-%s-%s, %d.", cycelem1[
           yr_elemx ], cycgendE[ yr_gender ], cycanim[ yr_animx ],
           target_year );
         fprintf ( fptgt, "%s\n", outbuf );
         sprintf ( outbuf, "Numeric square: %d.",
                   yr_smex );
         fprintf ( fptgt, "%s\n", outbuf );
       } // END - if ( inc_chin )
     else
       {
         sprintf ( outbuf, "Kalacakra New Year: %s, %d.", Prabhava[i],
                                target_year );
         fprintf ( fptgt, "\n%s\n", outbuf );
       }

    if ( inc_chin )
      start_month = first_Cmnth_flag;
    else
      start_month = first_Kmnth_flag;

// Start of month loop *******************************
    for ( mth_ind = start_month; mth_ind <= last_mnth_flag; ++mth_ind )
      {
// If first month, need Kalacakra new year data here:
        if ( inc_chin && newmon[mth_ind].mnthnumK == 1 )
          {
            sprintf ( outbuf, "Kalacakra New Year: %s, %d.", Prabhava[i],
                                   target_year );
            fprintf ( fptgt, "\n%s\n", outbuf );
          }

// Additional year information for second Chinese new year:

        if ( newmon[mth_ind].mnthnumK == 12
             && newyearintcal == 0
             && mth_ind > 6 && inc_chin ) // Second Chinese new year
          {
            newyearintcal = 1;
            if ( y_y < 7 )
              {
                yr_animx = ( y_y + 1200 - 7 ) % 12;
                yr_elemx = ( ( y_y + 1200 ) / 2 - 3 ) % 5;
                yr_gender = ( y_y + 1200 ) % 2; // 1 = female, odd; 0
              }
            else
              {
                yr_animx = ( y_y - 7 ) % 12;
                yr_elemx = ( y_y / 2 - 3 ) % 5;
                yr_gender = y_y % 2; // 1 = female, odd; 0 = male
              }
            yr_smex = 9 - ( ( y_y + 7 ) % 9 );

            sprintf ( outbuf, "Chinese New Year: %s-%s-%s, %d.",
              cycelem1[ yr_elemx ], cycgendE[ yr_gender ], cycanim[
              yr_animx ], y_y );
            fprintf ( fptgt, "\n%s\n", outbuf );
            sprintf ( outbuf, "Numeric square: %d.",
                      yr_smex );
            fprintf ( fptgt, "%s\n", outbuf );
          }
// End of additional year information. Now start month data:
        if ( inc_chin ) // calculate Chinese symbolic data:
          {
            // rabbit = 0, tiger = 11
            mt_animx = ( newmon[mth_ind].mnthnumK + 11 ) % 12;
            printf ( "Kmonth = %d, Cmonth = %d\n",
                  newmon[mth_ind].mnthnumK, mt_animx );
            Cmonth = newmon[mth_ind].mnthnumK + 1;
            if ( Cmonth > 12 )
              Cmonth -= 12;
            mt_elemx = yr_elemx + 1 + ( Cmonth - 1 ) / 2;
            mt_elemx = mt_elemx % 5;
// Remember there is another method for calculating these.
          }
// Start printing month information:
        if ( newmon[mth_ind].intercal == 1 ) // Intercalary
          {
            if ( inc_chin )
              {
                sprintf ( outbuf, "Intercalary month." );
                fprintf ( fptgt, "\n%s\n", outbuf );
                sprintf ( outbuf, "Chinese Month: %d, %s-%s, %s.", Cmonth,
                  cycelem1[ mt_elemx ], cycanim[ (int) mt_animx ],
                  M_season[ newmon[mth_ind].mnthnumK - 1 ] );
                fprintf ( fptgt, "%s\n", outbuf );
                sprintf ( outbuf, "Kalacakra Month: %d, %s.",
                      newmon[mth_ind].mnthnumK,
                      MnthS[ newmon[mth_ind].mnthnumK - 1 ] );
                fprintf ( fptgt, "%s\n\n", outbuf );
              }
            else
              {
                sprintf ( outbuf, "Month: %d, %s; Intercalary month.",
                      newmon[mth_ind].mnthnumK,
                      MnthS[ newmon[mth_ind].mnthnumK - 1 ] );
                fprintf ( fptgt, "\n%s\n\n", outbuf );
              }
          }
        else // Not intercalary
          {
            if ( inc_chin )
              {
                sprintf ( outbuf, "Chinese Month: %d, %s-%s, %s.", Cmonth,
                  cycelem1[ mt_elemx ], cycanim[ (int) mt_animx ],
                  M_season[ newmon[mth_ind].mnthnumK - 1 ] );
                fprintf ( fptgt, "\n%s\n", outbuf );
                sprintf ( outbuf, "Kalacakra Month: %d, %s, %s.",
                          newmon[mth_ind].mnthnumK,
                          MnthS[ newmon[mth_ind].mnthnumK - 1 ],
                          Zsigns[ newmon[ mth_ind ].Zsignend ] );
                fprintf ( fptgt, "%s\n\n", outbuf );
              }
            else
              {
                sprintf ( outbuf, "\nMonth: %d, %s, %s.\n\n",
                          newmon[mth_ind].mnthnumK,
                          MnthS[ newmon[mth_ind].mnthnumK - 1 ],
                          Zsigns[ newmon[ mth_ind ].Zsignend ] );
                fprintf ( fptgt, "%s", outbuf );
              }
          }

// End printing month information.
// Set date for the beginning of this new month:

        d_d = newmon[ mth_ind ].dd;
        m_m = newmon[ mth_ind ].mm;
        y_y = newmon[ mth_ind ].yy;
        more = 1;

// Now go through each day at a time:

        do
          { // Start of day loop
            // for each solar day, calculate the tithi(s)/lunar day.
            NFM_flag = 0; // Flag for new or full Moon days.
            SC_flag = 0; // Flag for Sun sign change.
            juldatf = jul_dat ( d_d, m_m, y_y );
            eclipse_flg = 0;
            juldat = gregd2jul ( (long int) d_d, (long int) m_m, (long int) y_y );

// Code here to work out day of week.
// Need to be careful of different time zones.

            dayofweek = ( long int ) floor ( juldatf );

            if ( d_bug )
              {
                sprintf ( outbuf, "\n\nDate: %d / %d / %d\n", d_d, m_m, y_y );
                fprintf ( fptgt_d, "%s\n", outbuf );
                sprintf ( outbuf, "Juldat = %ld\n", juldat );
                fprintf ( fptgt_d, "%s\n", outbuf );
              }
            dayofweek = juldat + 2L;
            dayofweek = dayofweek % 7L;

            if ( d_bug )
              {
                sprintf ( outbuf, "Julian UT 0: %f\n", juldatf );
                fprintf ( fptgt_d, "%s\n", outbuf );
                sprintf ( outbuf, "Time zone factor = %f\n", jul_factor );
                fprintf ( fptgt_d, "%s\n", outbuf );
              }

            juldatf = juldatf + jul_factor; // Add time zone factor.

            if ( d_bug )
              {
                sprintf ( outbuf, "Target time: %f\n", juldatf );
                fprintf ( fptgt_d, "%s\n", outbuf );
              }

            sun_mon ( juldatf ); // This adds Delta T
            moon_elong = mol - sol;

            if ( d_bug )
              {
                sprintf ( outbuf, "Solar longitude: %f\n", sol );
                fprintf ( fptgt_d, "%s\n", outbuf );
                sprintf ( outbuf, "Lunar longitude: %f\n", mol );
                fprintf ( fptgt_d, "%s\n", outbuf );
              }

            if ( moon_elong < 0.0 )
              moon_elong = moon_elong + 360.0;
            tt = 1 + (int) floor ( moon_elong / 12.0 );

            if ( d_bug )
              {
                sprintf ( outbuf, "Moon elongation: %f\n", moon_elong );
                fprintf ( fptgt_d, "%s\n", outbuf );
                sprintf ( outbuf, "Lunar date: %d\n", tt );
                fprintf ( fptgt_d, "%s\n", outbuf );
              }

// Check next lunar day, for omitted or duplicated:

            mol1 = mol; sol1 = sol; moon_elong1 = moon_elong;
            sun_mon ( juldatf + 1.0 ); // This adds Delta T
            mol2 = mol;
            sol2 = sol;
            moon_elong = mol - sol;
            if ( moon_elong < 0.0 )
              moon_elong = moon_elong + 360.0;
            moon_elong2 = moon_elong;
            tt2 = 1 + (int) floor ( moon_elong / 12.0 );
            if ( tt == tt2 )
              tt_lhag = 1;
            else
              {
                if ( tt2 < tt )
                  tt2 += 30;
                if ( tt2 - tt == 2 )
                  {
                    tt_chad = tt + 1;
                    if ( tt_chad > 30 )
                      tt_chad -= 30;
                  }
                else
                  tt_chad = 0;
              }

// Calculate here the time for tithi to change - if it does.
// This is all really only for test code, but may possibly
// be useful at some later point. Data in terms of Julian date

          if ( !tt_lhag ) // Then we have a change of tithi during the solar day
            {
              if ( !tt_chad ) // Normal day, only one change to find
                {
                  // Find the elongations needed:
                  tt1tgt = (double) tt * 12.0;
                  if ( tt == 30L )
                    tt1tgt = 0.0;
// Search from juldatf, the time of current daybreak
                  tt1chg = get_tithi ( tt1tgt, juldatf );
                }
              else // Two to find, omitted lunar date
                {
                  // Estimate the time of change for first tithi change:
                  tt1tgt = (double) tt * 12.0;
                  if ( tt == 30L )
                    tt1tgt = 0.0;
                  ott1chg = get_tithi ( tt1tgt, juldatf );
                  tt1chg = ott1chg; // Needed twice for test code.

                  // Now the next:
                  tt2tgt = (double) ( tt2 - 1L ) * 12.0;
                  if ( tt2 - 1L == 30L )
                    tt2tgt = 0.0;
                  else if ( tt2 - 1L == 31L )
                    tt2tgt = 12.0;
                  tt2chg = get_tithi ( tt2tgt, juldatf );
                }
            }
// Second of duplicated, tithi will change:
          else if ( tt_lhag == 2 )
            {
              // Find the elongation needed:
              tt1tgt = (double) tt * 12.0;
              if ( tt == 30L )
                tt1tgt = 0.0;
              tt1chg = get_tithi ( tt1tgt, juldatf );
            }

// Put back the data:
          mol = mol1; sol = sol1; moon_elong = moon_elong1;

// Calculate lunar mansion at daybreak:

          lunmanx = (int) floor ( mol * 0.075 ); // 3/40;

          if ( d_bug )
            {
              sprintf ( outbuf, "Lunar mansion: %d\n", lunmanx );
              fprintf ( fptgt_d, "%s\n", outbuf );
            }

// Calculate time of lunar mansion change:

          mol1 = mol; sol1 = sol;
          tgtmol = (double) (lunmanx + 1) * 13.33333333; // 40/3
          if ( lunmanx == 26 )
            {
              tgtmol = 360.0;
              lunmannext = 0;
            }
          else
            lunmannext = lunmanx + 1;
          lmchgjdatf = juldatf;

          do {
               lmchgjdatf = lmchgjdatf + ( tgtmol - mol ) / 13.2;
               sun_mon ( lmchgjdatf );
               if ( mol < 6.0 )
                 mol = mol + 360.0;
             } while ( tgtmol - mol > 0.000001 || mol - tgtmol > 0.000001 );

          if ( lmchgjdatf >= juldatf + 1.0 )
            lmchg = 0;
          else
            {
              lmchg = 1;
              dif = 60.0 * ( lmchgjdatf - juldatf ); // Seems right.
              lmcn = (int) floor ( dif );
              dif = 60.0 * ( dif - (double) lmcn );
              lmcp = (int) floor ( dif );
              dif = 6.0 * ( dif - (double) lmcp );
              lmcb = (int) floor ( dif );
            }

// Put back the data:
          mol = mol1; sol = sol1;

// Calculate yoga:

// The following addition sets for the Tsurphu tradition of the indexing of the yogas.
// For the Phugpa interpretation, do not add 346.6666...
          x = mol + sol + 346.666667;
          if ( x >= 360.0 )
            x = x - 360.0;
          if ( x >= 360.0 )
            x = x - 360.0;
          sbyorba = (int) floor ( ( x * 3.0 ) / 40.0 );

// Calculate karana:

          byedpa = (int) floor ( moon_elong / 6.0 );

          if ( byedpa == 0 )
            byedpax = 7;
          else if ( byedpa == 57 )
            byedpax = 8;
          else if ( byedpa == 58 )
            byedpax = 9;
          else if ( byedpa == 59 )
            byedpax = 10;
          else
            byedpax = ( byedpa - 1 ) % 7;

// Now, calculate chinese symbolic information:
// Solar day data:

          if ( inc_chin )
            {
              sd_animx = (int) ( ( juldat - 2L ) % 12L );
              sd_elemx = (int) ( ( ( juldat - 3L ) / 2L ) % 5L );
              sd_clunmanx = (int) ( ( juldat - 17L ) % 28L );
              sd_smex = 9 - (int) ( juldat % 9L );
              sd_parx = (int) ( ( juldat - 5713L + 3L ) % 8 );

// Lunar day data:

// If Chinese month is number 1, Trigram is Li, index = 1
              ld_parx = ( Cmonth - 1 ) * 30 + tt;
              ld_parx = ld_parx % 8;
// If Chinese month is number 1, "sme ba" is 1
              ld_smex = ( Cmonth - 1 ) * 30 + tt;
              ld_smex = ld_smex % 9;
              if ( ld_smex == 0 )
                ld_smex = 9;
// If Chinese month is number 1, Animal is Tiger, index = 11
              ld_elemx = ( mt_elemx + tt ) % 5;
              ld_animx = ( Cmonth - 1 ) * 30 + tt + 10;
              ld_animx = ld_animx % 12;
// From "Utngon" text, cycling planet:
// Text is: 'byung rtsis utpala sngon po
              if ( Cmonth == 1 || Cmonth == 5 || Cmonth == 9 )
                ld_planx = 0;
              else if ( Cmonth == 2 || Cmonth == 6 || Cmonth == 10 )
                ld_planx = 2;
              else if ( Cmonth == 3 || Cmonth == 7 || Cmonth == 11 )
                ld_planx = 3;
              else if ( Cmonth == 4 || Cmonth == 8 || Cmonth == 12 )
                ld_planx = 5;
              ld_planx = ld_planx + tt - 1;
              ld_planx = ld_planx % 7;
// From "Utngon" text, cycling lunar mansion:
              if ( Cmonth == 1 ) ld_lunmx = 22;
                else if ( Cmonth == 2 ) ld_lunmx = 24;
              else if ( Cmonth == 3 ) ld_lunmx = 26;
              else if ( Cmonth == 4 ) ld_lunmx = 1;
              else if ( Cmonth == 5 ) ld_lunmx = 3;
              else if ( Cmonth == 6 ) ld_lunmx = 4;
              else if ( Cmonth == 7 ) ld_lunmx = 8;
              else if ( Cmonth == 8 ) ld_lunmx = 10;
              else if ( Cmonth == 9 ) ld_lunmx = 12;
              else if ( Cmonth == 10 ) ld_lunmx = 15;
              else if ( Cmonth == 11 ) ld_lunmx = 17;
              else if ( Cmonth == 12 ) ld_lunmx = 19;
              ld_lunmx = ld_lunmx + tt - 15;
              if ( ld_lunmx < 0 )
                ld_lunmx = ld_lunmx + 28;
              else if ( ld_lunmx > 27 )
                ld_lunmx = ld_lunmx - 28;
            } // END - if ( inc_chin )

// Check for eclipses:

            if ( tt_lhag == 0 && tt_chad != 15 && tt == 15 ) // Normal day
              eclipse_flg = chk_eclipse ( juldatf, 15 );
            else if ( tt_lhag == 2 && tt == 15 ) // Duplicated day
              eclipse_flg = chk_eclipse ( juldatf, 15 );
            else if ( tt_chad == 15 ) // Omitted day
              eclipse_flg = chk_eclipse ( juldatf, 15 );
            else if ( tt_lhag == 0 && tt_chad != 30 && tt == 30 ) // Normal day
              eclipse_flg = chk_eclipse ( juldatf, 30 );
            else if ( tt_lhag == 2 && tt == 30 ) // Duplicated day
              eclipse_flg = chk_eclipse ( juldatf, 30 );
            else if ( tt_chad == 30 ) // Omitted day
              eclipse_flg = chk_eclipse ( juldatf, 30 );

// Now, start printing daily information, first,
// dealing with omitted lunar days:

            if ( tt - last_tt > 1 )  // Omitted Tithi
                                     // Omitted 30th is dealt with later
              {
// LINE 1:
                sprintf ( outbuf, "%d. Omitted lunar day.", tt-1 );
                fprintf ( fptgt, "%s\n", outbuf );

// LINE 1a: -- Debug code, not normally in use:

//            sprintf ( outbuf, "Tithi 1: %f, tithi 2: %f.", ott1chg, tt2chg );
//            fprintf ( fptgt, "%s\n", outbuf );

// LINE 2: (if special)
                if ( !newmon[mth_ind].intercal &&
                  chk_spec ( newmon[mth_ind].mnthnumK, tt-1 ) )
                    {
                      fprintf ( fptgt, "\x20\x20\x20\x20%s\n", outbuf );
                    }
                if ( inc_chin )
                  {
// LINE 2:
// This line left out of earlier versions of the calendar. It is not
// particularly important, but Tibetan calendars do contain this information:
// tt-1 is the omitted lunar date

// If Chinese month is number 1, Trigram is Li, index = 1
                    ldo_parx = ( Cmonth - 1 ) * 30 + tt - 1;
                    ldo_parx = ldo_parx % 8;
// If Chinese month is number 1, "sme ba" is 1  // Kham
                    ldo_smex = ( Cmonth - 1 ) * 30 + tt - 1;
                    ldo_smex = ldo_smex % 9;
                    if ( ldo_smex == 0 )
                      ldo_smex = 9;
// If Chinese month is number 1, Animal is Tiger, index = 11
                    ldo_animx = ( Cmonth - 1 ) * 30 + tt + 9;
                    ldo_animx = ldo_animx % 12;
                    sprintf ( outbuf, "%s, %s %d", cycanim[ ldo_animx ],
                              cycpar[ldo_parx], ldo_smex );
                    fprintf ( fptgt, "%s\n", outbuf );
// LINE 3:

// Now, zin phung, yan kwong, etc:
                    chk_sadag ( newmon[mth_ind].mnthnumK, tt-1 );
                    if ( yan_kwong_flg || zin_phung_flg )
                      {
                        if ( yan_kwong_flg && zin_phung_flg )
                          sprintf ( outbuf, "\x20\x20zin phung, yan kwong" );
                        else if ( zin_phung_flg )
                          sprintf ( outbuf, "\x20\x20zin phung" );
                        else if ( yan_kwong_flg )
                          sprintf ( outbuf, "\x20\x20yan kwong" );
                        fprintf ( fptgt, "\x20\x20%s\n", outbuf );
                      }
                  } // END - if ( inc_chin )
              } // END - if ( tt - last_tt > 1 )  // Omitted Tithi

// Now, normal, duplicated or omitted 30th:

// LINE 1:

            if ( lmchg ) // If lunar mansion changes:
              {
            sprintf ( outbuf, "%d. %s, %d %s %d. %s, %d:%d:%d: %s; %s, %s.",
            tt, dayoweek[ (int) dayofweek ], d_d, wmonths[m_m-1],
            y_y, lunmanP[ lunmanx ], lmcn, lmcp, lmcb,
            lunmanP[ lunmannext ], yogaP[ sbyorba ], byedP[ byedpax ] );
              }
            else
              {
            sprintf ( outbuf, "%d. %s, %d %s %d. %s; %s, %s.",
                      tt, dayoweek[ (int) dayofweek ], d_d, wmonths[m_m-1],
                       y_y, lunmanP[ lunmanx ], yogaP[ sbyorba ],
                      byedP[ byedpax ] );
              }
            fprintf ( fptgt, "%s\n", outbuf );

// LINE 1B:
            if ( lmchg ) // If lunar mansion changes:
              {
                phrod1 = getphrod ((int) dayofweek, lunmanx, mol1 );
                phrod2 = getphrod ((int) dayofweek, lunmannext, mol1 );
                sprintf ( outbuf, "%s-%s -> %s-%s, %s -> %s.",
                          wdayelem[(int) dayofweek], lmanelem[lunmanx],
                          wdayelem[(int) dayofweek], lmanelem[lunmannext],
                          phrodE[phrod1], phrodE[phrod2] );
              }
            else
              {
                phrod1 = getphrod ((int) dayofweek, lunmanx, mol1 );
                sprintf ( outbuf, "%s-%s, %s.",
                          wdayelem[(int) dayofweek], lmanelem[lunmanx],
                          phrodE[phrod1] );
              }
            fprintf ( fptgt, "\x20\x20%s\n", outbuf );

// LINE 1c: -- Debug code, not normally in use:

//        if ( tt_lhag != 1 )
//          {
//            sprintf ( outbuf, "Tithi: %f", tt1chg );
//            fprintf ( fptgt, "%s\n", outbuf );
//          }
            
// LINE 2:
            if ( inc_chin ) // Changed 1/9/2004
              {
                sprintf ( outbuf, "%s-%s, %s %d %s %s", cycelem1[ ld_elemx ],
                          cycanim[ld_animx], cycpar[ld_parx], ld_smex,
                          dayoweek[ld_planx], Clunman[ld_lunmx] );
                fprintf ( fptgt, "\x20\x20%s\n", outbuf );
// LINE 3:
                sprintf ( outbuf, "Solar: %s-%s. %s, %s %d",
                          cycelem1[sd_elemx], cycanim[sd_animx],
                          Clunman[sd_clunmanx], cycpar[sd_parx], sd_smex );
                fprintf ( fptgt, "\x20\x20%s\n", outbuf );
              }

// Extra lines:
            if ( ( ( tt == 14 && tt_chad == 15 ) ||
                ( tt == 19 && tt_chad == 30 ) ) && eclipse_flg )
              {
                fprintf ( fptgt, "\x20\x20\x20\x20%s\n", eclps_str );
              }

// We need to skip this if repeated lunar day:

            if ( tt_lhag != 2 )
              {
//NOW, zin phung, yan kwong, etc:
                if ( inc_chin )
                  {
                    chk_sadag ( newmon[mth_ind].mnthnumK, tt );
                    if ( yan_kwong_flg || zin_phung_flg )
                      {
                        if ( yan_kwong_flg && zin_phung_flg )
                          sprintf ( outbuf, "\x20\x20zin phung, yan kwong" );
                        else if ( zin_phung_flg )
                          sprintf ( outbuf, "\x20\x20zin phung" );
                        else if ( yan_kwong_flg )
                          sprintf ( outbuf, "\x20\x20yan kwong" );
                        fprintf ( fptgt, "\x20\x20%s\n", outbuf );
                      }
                  }
              }

// We need to skip this if repeated lunar day:

            if ( !newmon[mth_ind].intercal && ( tt_lhag != 2 ) &&
                 chk_spec ( newmon[mth_ind].mnthnumK, tt ) )
              fprintf ( fptgt, "\x20\x20\x20\x20%s\n", outbuf );

            if ( chk_sign ( juldatf, sol ) )  // Check for change of sign
              // Result is put into outbufSC ready for printing.
              SC_flag = 1; // Flag for Sun sign change.

// This only works if 15 or 30 are not omitted days. This should be fixed:
            if ( ( tt == 15 || tt == 30 ) && tt_lhag != 1 )
              {
                NFM_flag = 1; // Flag for new or full Moon days.
                if ( !get_NFMoon ( juldatf, tt ) ) // Error code:
                  {
                    printf ( "Real problem - NO New/Full Moon found\n" );
                    getch ();
                  }
              }

            if ( SC_flag && NFM_flag )
              {
                if ( jul_SC < jul_NFM ) // Sign change goes first:
                  {
                    fprintf ( fptgt, "\x20\x20%s\n", outbufSC );
                    fprintf ( fptgt, "\x20\x20\x20\x20%s\n", outbufNFM );
                  }
                else // New/Full Moon goes first.
                  {
                    fprintf ( fptgt, "\x20\x20\x20\x20%s\n", outbufNFM );
                    fprintf ( fptgt, "\x20\x20%s\n", outbufSC );
                  }
              }
            else  // If only one needs to be printed.
              {
                if ( SC_flag )
                  {
                    fprintf ( fptgt, "\x20\x20%s\n", outbufSC );
                  }
                else if ( NFM_flag )
                  {
                    fprintf ( fptgt, "\x20\x20\x20\x20%s\n", outbufNFM );
                  }
              }

            if ( NFM_flag && tt_lhag != 1 && eclipse_flg )
              {
                fprintf ( fptgt, "\x20\x20\x20\x20%s\n", eclps_str );
              }
            if ( tt_lhag == 2 )
              {
                fprintf ( fptgt, "\x20\x20%s\n", "Duplicated lunar day." );
                tt_lhag = 0;
              }
            else if ( tt_lhag == 1 )
              tt_lhag = 2;

// This is for omitted 30th:
// Best to leave it here as it comes at the end of the month.

            chg_tim ( 0, 0, 1, 0, 0 );
            if ( tt == 29 && d_d == newmon[ mth_ind+1 ].dd )
              {
// LINE 1:
                sprintf ( outbuf, "%d. Omitted lunar day.", 30 );
                fprintf ( fptgt, "%s\n", outbuf );
                more = 0;
// LINE 1B:
                if ( lmchg ) // If lunar mansion changes:
                 {
                   phrod1 = getphrod ((int) dayofweek, lunmanx, mol1 );
                   phrod2 = getphrod ((int) dayofweek, lunmannext, mol1 );
                   sprintf ( outbuf, "%s-%s -> %s-%s, %s -> %s.",
                          wdayelem[(int) dayofweek], lmanelem[lunmanx],
                          wdayelem[(int) dayofweek], lmanelem[ lunmannext ],
                          phrodE[phrod1], phrodE[phrod2] );
                 }
                else
                 {
                   phrod1 = getphrod ((int) dayofweek, lunmanx, mol1 );
                   sprintf ( outbuf, "%s-%s, %s.",
                             wdayelem[(int) dayofweek], lmanelem[lunmanx],
                             phrodE[phrod1] );
                 }
               fprintf ( fptgt, "\x20\x20%s\n", outbuf );

// LINE 2:
                if ( inc_chin )
                  {
// tt+1 is now the omitted lunar date
                    ldo_parx = ( Cmonth - 1 ) * 30 + tt + 1;
                    ldo_parx = ldo_parx % 8;
                    ldo_smex = ( Cmonth - 1 ) * 30 + tt + 1;
                    ldo_smex = ldo_smex % 9;
                    if ( ldo_smex == 0 )
                      ldo_smex = 9;
                    ldo_animx = ( Cmonth - 1 ) * 30 + tt + 11;
                    ldo_animx = ldo_animx % 12;
                    sprintf ( outbuf, "%s, %s %d", cycanim[ ldo_animx ],
                              cycpar[ ldo_parx ], ldo_smex );
                    fprintf ( fptgt, "\x20\x20%s\n", outbuf );
                  }
              }
            if ( tt == 30 && d_d == newmon[ mth_ind+1 ].dd )
              more = 0;
            last_tt = tt;
          } while ( more ); // END OF DAY LOOP - end of day loop

        last_tt = 0;
      } // END OF MONTH LOOP.
    fclose ( fptgt );
  } // END - prn_cal

// juldat is set to juldatf = time at daybreak at target location.
// Checking for time of new/full Moon. sl = solar, ml = lunar longitudes.
int get_NFMoon ( double juldat, int tt )
  {
    double jd, /*tgt,*/ dif, melong;
    int h, m, s, n, p, b;
    int done = 0;

    sun_mon ( juldat ); // This adds Delta T
// This only works if 15 or 30 are not omitted days. This should be fixed.
    if ( tt == 30 ) // New moon
      {
        melong = mol - sol;
        if ( melong < -330.0 )
          melong = melong + 360.0;
        else if ( melong > 330.0 )
          melong = melong - 360.0;
        jd = juldat;
        do
          {
            jd = jd - ( melong / 12.0 ); // Mean  Moon moves 13.2 degrees/day
            if ( jd < juldat || jd > (juldat + 1.0) )
              {
                printf ( "Problem 2 searching for new Moon.\n" );
                getch ();
              }
            sun_mon ( jd );
            melong = mol - sol;

            if ( melong > -0.00001 && melong < 0.00001 )
              done = 1;
            if ( melong < -330.0 )
              melong = melong + 360.0;
            else if ( melong > 330.0 )
              melong = melong - 360.0;
            if ( melong < -15.0 || melong > 15.0 )
              {
                printf ( "ERROR - Sun: %f, Moon: %f\n", sol, mol );
                getch ();
                return (0);
              }
          } while ( !done );

        jul_NFM = jd;
        dif = 60.0 * ( jd - juldat );  // Seems right.
        n = (int) floor ( dif );
        dif = 60.0 * ( dif - (double) n );
        p = (int) floor ( dif );
        dif = 6.0 * ( dif - (double) p );
        b = (int) floor ( dif );
// Test code for checking accuracy, commented out:
//      printf ("\nDate: %d:%d:%d\n", d_d, m_m, y_y );
        sprintf ( outbufNFM, "New Moon at %d:%d:%d.", n, p, b );
//      printf ( "n:p:b: %s\n", outbufNFM );
// Convert nadi, pala and breaths from daybreak,
// into hours, minutes and seconds from midnight:
//      conv2hms ( &h, &m, &s, n, p, b );
//      sprintf ( outbufNFM, "New Moon at %d:%d:%d.", h, m, s );
//      printf ( "h:m:s: %s\n", outbufNFM );
//      getch ();
        return (1);
      }
    else if ( tt == 15 ) // Full moon
      {
        melong = mol - sol;
        if ( melong < 0.0 )
          melong = melong + 360.0;
        jd = juldat;
        do
          {
            jd = jd - ( ( melong - 180.0 ) / 12.0 ); // Mean Moon moves 13.2
                                                     // degrees/day
            if ( jd < juldat || jd > (juldat + 1.0) )
              {
                printf ( "Problem 2 searching for full Moon.\n" );
                getch ();
              }
            sun_mon ( jd );
            melong = mol - sol;
            if ( melong < 0.0 )
              melong = melong + 360.0;

            if ( melong > 179.99999 && melong < 180.00001 )
              done = 1;
            if ( melong < 164.0 || melong > 196.0 )
              {
                printf ( "ERROR - Sun: %f, Moon: %f\n", sol, mol );
                getch ();
                return (0);
              }
          } while ( !done );

        jul_NFM = jd;
        dif = 60.0 * ( jd - juldat );  // Seems right.
        n = (int) floor ( dif );
        dif = 60.0 * ( dif - (double) n );
        p = (int) floor ( dif );
        dif = 6.0 * ( dif - (double) p );
        b = (int) floor ( dif );
        
// Test code for checking accuracy, commented out:
//      printf ("\nDate: %d:%d:%d\n", d_d, m_m, y_y );
        sprintf ( outbufNFM, "Full Moon at %d:%d:%d.", n, p, b );
//      printf ( "n:p:b: %s\n", outbufNFM );
// Convert nadi, pala and breaths from daybreak,
// into hours, minutes and seconds from midnight:
//      conv2hms ( &h, &m, &s, n, p, b );
//      sprintf ( outbufNFM, "Full Moon at %d:%d:%d.", h, m, s );
//      printf ( "h:m:s: %s\n", outbufNFM );        
        return (1);
      }
    return (0); // 0 is error code here, no new or full Moon found.
  } // END - get_NFMoon ()

double getmeanSun ( double juld )
  {
    double T, TM, TM2, TM3, TM4, TM5, X, delt, SunL;

// From sun_mon:
    T = (juld - 2451545.0);
    delt = get_deltaT (juld) / 86400.0; // This converts seconds to fraction of a day
    T = (T + delt) / 36525.0;  // This is now in Julian centuries.

    TM = T / 10.0; // Measured in Julian millenia
    TM2 = TM * TM;
    TM3 = TM * TM2;
    TM4 = TM2 * TM2;
    TM5 = TM * TM4;

// Mean Sun, Astronomical Algorithms, p.183: Should compare with p.212

    X = 280.4664567 + 360007.6982779 * TM;
    X = modfeh ( X, 360.0 );
    SunL = X + 0.03032028 * TM2
             + TM3 / 49931.0
             - TM4 / 15300.0
             - TM5 / 2000000.0;
    SunL = modfeh ( SunL, 360.0 );
    if ( SunL < 0.0 )
      SunL = SunL + 360.0;
    return ( SunL );
  } // END - getmeanSun

int getphrod ( int dow, int man, double lunlong )
  {
    int x;
// First cater for the difference between 27 & 28 lunar mansion sets
    if ( man > 21 )
      {
        ++man;
      }
    else if ( man == 21 )
      {
        if ( lunlong > 286.666666 )
          man = 22;
      }
    x = man + 32;
    x = x - 4 * dow;
    if ( x < 0 )
      {
        printf ( "ERROR - phrod zero!\n" );
        getch ();
      }
    if ( x > 27 )
      x = x - 28;
    if ( x > 27 )
      x = x - 28;
    return (x);
  } // END - getphrod ()

// This converts Kalacakra time notation in nadi, pala and breaths, from daybreak
// into standard hours, minutes and seconds, from midnight.
void conv2hms ( int *h, int *m, int *s, int n, int p, int b )
  {
    double tim;
    tim = (( (double) n * 60.0 + (double) p ) * 6.0 + (double) b ) / 21600.0;
    tim = tim * 24.0;
    *h = (int) floor (tim) + 5; // Convert to LMST
    tim = tim - floor (tim);
    tim = tim * 60.0;
    *m = (int) floor (tim);
    tim = tim - floor (tim);
    tim = tim * 60.0;
    *s = (int) floor (0.5 + tim);
  } // END - conv2hms ()
